#!/usr/bin/env python
# tool to materialize feature_sources and feature_sets into thrift configurations
# that chronon jobs can consume


#     Copyright (C) 2023 The Chronon Authors.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import logging
import os

import click

import ai.chronon.api.ttypes as api
import ai.chronon.repo.extract_objects as eo
import ai.chronon.utils as utils
from ai.chronon.api.ttypes import GroupBy, Join, StagingQuery
from ai.chronon.repo import JOIN_FOLDER_NAME, \
    GROUP_BY_FOLDER_NAME, STAGING_QUERY_FOLDER_NAME, TEAMS_FILE_PATH
from ai.chronon.repo import teams
from ai.chronon.repo.serializer import thrift_simple_json_protected
from ai.chronon.repo.validator import ChrononRepoValidator

# This is set in the main function -
# from command line or from env variable during invocation
FOLDER_NAME_TO_CLASS = {
    GROUP_BY_FOLDER_NAME: GroupBy,
    JOIN_FOLDER_NAME: Join,
    STAGING_QUERY_FOLDER_NAME: StagingQuery,
}

DEFAULT_TEAM_NAME = "default"


def get_folder_name_from_class_name(class_name):
    return {v.__name__: k for k, v in FOLDER_NAME_TO_CLASS.items()}[class_name]


@click.command()
@click.option(
    '--chronon_root',
    envvar='CHRONON_ROOT',
    help='Path to the root chronon folder',
    default=os.getcwd())
@click.option(
    '--input_path', '--conf', 'input_path',
    help='Relative Path to the root chronon folder, which contains the objects to be serialized',
    required=True)
@click.option(
    '--output_root',
    help='Relative Path to the root chronon folder, to where the serialized output should be written',
    default="production")
@click.option(
    '--debug',
    help='debug mode',
    is_flag=True)
@click.option(
    '--force-overwrite',
    help='Force overwriting existing materialized conf.',
    is_flag=True)
def extract_and_convert(chronon_root, input_path, output_root, debug, force_overwrite):
    """
    CLI tool to convert Python chronon GroupBy's, Joins and Staging queries into their thrift representation.
    The materialized objects are what will be submitted to spark jobs - driven by airflow, or by manual user testing.
    """
    if debug:
        log_level = logging.DEBUG
    else:
        log_level = logging.INFO
    _print_highlighted("Using chronon root path", chronon_root)
    chronon_root_path = os.path.expanduser(chronon_root)
    path_split = input_path.split('/')
    obj_folder_name = path_split[0]
    obj_class = FOLDER_NAME_TO_CLASS[obj_folder_name]
    full_input_path = os.path.join(chronon_root_path, input_path)
    _print_highlighted(f"Input {obj_folder_name} from", full_input_path)
    assert os.path.exists(full_input_path), f"Input Path: {full_input_path} doesn't exist"
    if os.path.isdir(full_input_path):
        results = eo.from_folder(chronon_root_path, full_input_path, obj_class, log_level=log_level)
    elif os.path.isfile(full_input_path):
        assert full_input_path.endswith(".py"), f"Input Path: {input_path} isn't a python file"
        results = eo.from_file(chronon_root_path, full_input_path, obj_class, log_level=log_level)
    else:
        raise Exception(f"Input Path: {full_input_path}, isn't a file or a folder")
    validator = ChrononRepoValidator(chronon_root_path, output_root, log_level=log_level)
    extra_online_group_bys = {}
    num_written_objs = 0
    full_output_root = os.path.join(chronon_root_path, output_root)
    teams_path = os.path.join(chronon_root_path, TEAMS_FILE_PATH)
    for name, obj in results.items():
        team_name = name.split(".")[0]
        _set_team_level_metadata(obj, teams_path, team_name)
        _set_templated_values(obj, obj_class, teams_path, team_name)
        if _write_obj(full_output_root, validator, name, obj, log_level, force_overwrite, force_overwrite):
            num_written_objs += 1

            # In case of online join, we need to materialize the underlying online group_bys.
            if obj_class is Join and obj.metaData.online:
                online_group_bys = {}
                offline_gbs = []  # gather list to report errors
                for jp in obj.joinParts:
                    if jp.groupBy.metaData.online:
                        online_group_bys[jp.groupBy.metaData.name] = jp.groupBy
                    else:
                        offline_gbs.append(jp.groupBy.metaData.name)
                extra_online_group_bys.update(online_group_bys)
                assert not offline_gbs, \
                    "You must make all dependent GroupBys `online` if you want to make your join `online`." \
                    " You can do this by passing the `online=True` argument to the GroupBy constructor." \
                    " Fix the following: {}".format(offline_gbs)
    if extra_online_group_bys:
        num_written_group_bys = 0
        # load materialized joins to validate the additional group_bys against.
        validator.load_objs()
        for name, obj in extra_online_group_bys.items():
            team_name = name.split(".")[0]
            _set_team_level_metadata(obj, teams_path, team_name)
            if _write_obj(full_output_root, validator, name, obj, log_level,
                          force_compile=True, force_overwrite=force_overwrite):
                num_written_group_bys += 1
        print(f"Successfully wrote {num_written_group_bys} online GroupBy objects to {full_output_root}")
    if num_written_objs > 0:
        print(f"Successfully wrote {num_written_objs} {(obj_class).__name__} objects to {full_output_root}")


def _set_team_level_metadata(obj: object, teams_path: str, team_name: str):
    namespace = teams.get_team_conf(teams_path, team_name, "namespace")
    table_properties = teams.get_team_conf(teams_path, team_name, "table_properties")
    obj.metaData.outputNamespace = obj.metaData.outputNamespace or namespace
    obj.metaData.tableProperties = obj.metaData.tableProperties or table_properties
    obj.metaData.team = team_name


def __fill_template(table, obj, namespace):
    if table:
        table = table.replace('{{ logged_table }}', utils.log_table_name(obj, full_name=True))
        table = table.replace('{{ db }}', namespace)
    return table


def _set_templated_values(obj, cls, teams_path, team_name):
    namespace = teams.get_team_conf(teams_path, team_name, "namespace")
    if cls == api.Join and obj.bootstrapParts:
        for bootstrap in obj.bootstrapParts:
            bootstrap.table = __fill_template(bootstrap.table, obj, namespace)
        if obj.metaData.dependencies:
            obj.metaData.dependencies = [__fill_template(dep, obj, namespace) for dep in obj.metaData.dependencies]
    if cls == api.Join and obj.labelPart:
        obj.labelPart.metaData.dependencies = [label_dep.replace('{{ join_backfill_table }}',
                                                                 utils.output_table_name(obj, full_name=True))
                                               for label_dep in obj.labelPart.metaData.dependencies]


def _write_obj(full_output_root: str,
               validator: ChrononRepoValidator,
               name: str,
               obj: object,
               log_level: int,
               force_compile: bool = False,
               force_overwrite: bool = False) -> bool:
    """
    Returns True if the object is successfully written.
    """
    team_name = name.split(".")[0]
    obj_class = type(obj)
    class_name = obj_class.__name__
    name = name.split('.', 1)[1]
    _print_highlighted(f"{class_name} Team", team_name)
    _print_highlighted(f"{class_name} Name", name)
    obj_folder_name = get_folder_name_from_class_name(class_name)
    output_path = os.path.join(full_output_root, obj_folder_name, team_name)
    output_file = os.path.join(output_path, name)
    skip_reasons = validator.can_skip_materialize(obj)
    if not force_compile and skip_reasons:
        reasons = ', '.join(skip_reasons)
        _print_warning(f"Skipping {class_name} {name}: {reasons}")
        if os.path.exists(output_file):
            _print_warning(f"old file exists for skipped config: {output_file}")
        return False
    validation_errors = validator.validate_obj(obj)
    if validation_errors:
        _print_error(f"Could not write {class_name} {name}",
                     ', '.join(validation_errors))
        return False
    if force_overwrite:
        _print_warning(f"Force overwrite {class_name} {name}")
    elif not validator.safe_to_overwrite(obj):
        _print_warning(f"Cannot overwrite {class_name} {name} with existing online conf")
        return False
    _write_obj_as_json(name, obj, output_file, obj_class)
    return True


def _write_obj_as_json(name: str, obj: object, output_file: str, obj_class: type):
    class_name = obj_class.__name__
    output_folder = os.path.dirname(output_file)
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)
    assert os.path.isdir(output_folder), f"{output_folder} isn't a folder."
    assert (hasattr(obj, "name") or hasattr(obj, "metaData")), \
        f"Can't serialize objects without the name attribute for object {name}"
    with open(output_file, "w") as f:
        _print_highlighted(f"Writing {class_name} to", output_file)
        f.write(thrift_simple_json_protected(obj, obj_class))


def _print_highlighted(left, right):
    # print in blue.
    print(f"{left:>25} - \u001b[34m{right}\u001b[0m")


def _print_error(left, right):
    # print in red.
    print(f"\033[91m{left:>25} - \033[1m{right}\033[00m")


def _print_warning(string):
    # print in yellow - \u001b[33m
    print(f"\u001b[33m{string}\u001b[0m")


if __name__ == '__main__':
    extract_and_convert()



#!/usr/bin/env python3


#     Copyright (C) 2023 The Chronon Authors.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from contextlib import contextmanager
from pathlib import Path

import argparse
import json
import os
import subprocess


CWD = os.getcwd()
GB_INDEX_SPEC = {
    "sources": [
        "sources[].events.table",
        "sources[].entities.snapshotTable",
        "sources[].entities.mutationTable",
        "sources[].entities.topic",
        "sources[].events.topic",
    ],
    "_event_tables": ["sources[].events.table"],
    "_event_topics": ["sources[].events.topic"],
    "aggregation": [
        "aggregations[].inputColumn"
    ],
    "keys": [
        "keyColumns"
    ],
    "name": [
        "metaData.name"
    ],
    "online": [
        "metaData.online"
    ],
    "output_namespace": [
        "metaData.outputNamespace"
    ],
}

JOIN_INDEX_SPEC = {
    "input_table": [
        "left.entities.snapshotTable",
        "left.events.table",
    ],
    "_events_driver": ["left.events.table"],
    "group_bys": [
        "joinParts[].groupBy.metaData.name",
        "rightParts[].groupBy.name",
    ],
    "name": [
        "metaData.name"
    ],
    "output_namespace": [
        "metaData.outputNamespace"
    ],
    "_group_bys": [
        "joinParts[].groupBy",
        "rightParts[].groupBy"
    ]
}

DEFAULTS_SPEC = {
    "output_namespace": "namespace",
}

GB_REL_PATH = "production/group_bys"
JOIN_REL_PATH = "production/joins"
FILTER_COLUMNS = ["aggregation", "keys", "name", "sources", "joins"]
PATH_FIELDS = ['file', 'json_file']
# colors chosen to be visible clearly on BOTH black and white terminals
# change with caution
NORMAL = '\033[0m'
BOLD = '\033[1m'
ITALIC = '\033[3m'
UNDERLINE = '\033[4m'
RED = '\033[38;5;160m'
GREEN = '\033[38;5;28m'
ORANGE = '\033[38;5;130m'
BLUE = '\033[38;5;27m'
GREY = '\033[38;5;246m'
HIGHLIGHT = BOLD+ITALIC+RED


# walks the json nodes recursively collecting all values that match the path
# a trailing `[]` in a field in the path indicates that there is an array of
# object in the correspoding node value.
def extract_json(json_path, conf_json):
    if json_path is None:
        return conf_json
    steps = json_path.split(".", 1)
    key = steps[0]
    next = steps[1] if len(steps) > 1 else None
    if key.endswith("[]"):
        key = key[:-2]
        if key in conf_json:
            result = []
            for value in conf_json[key]:
                result.extend(extract_json(next, value))
            return result
    else:
        if key in conf_json:
            final = extract_json(next, conf_json[key])
            if isinstance(final, list):
                return final
            else:
                return [final]
    return []


def build_entry(conf, index_spec, conf_type, root=CWD, teams=None):
    conf_dict = conf
    if isinstance(conf, str):
        with open(conf) as conf_file:
            try:
                conf_dict = json.load(conf_file)
            except BaseException as ex:
                print(f"Failed to parse {conf} due to :: {ex}")
                return
    entry = {"file": None}
    for column, paths in index_spec.items():
        result = []
        for path in paths:
            result.extend(extract_json(path, conf_dict))
        entry[column] = result

    if len(entry["name"]) == 0:
        return None

    # derive python file path from the name & conf_type
    (team, conf_module) = entry["name"][0].split(".", 1)
    # Update missing values with teams defaults.
    for field, mapped_field in DEFAULTS_SPEC.items():
        if field in entry and not entry[field]:
            entry[field] = [teams[team][mapped_field]]

    file_base = "/".join(conf_module.split(".")[:-1])
    py_file = file_base + ".py"
    init_file = file_base + "/__init__.py"
    py_path = os.path.join(root, conf_type, team, py_file)
    init_path = os.path.join(root, conf_type, team, init_file)
    conf_path = py_path if os.path.exists(py_path) else init_path
    entry["json_file"] = os.path.join(root, "production", conf_type, team, conf_module)
    entry["file"] = conf_path
    return entry


@contextmanager
def chdir(path):
    """
    Context manager to run subprocesses in the appropriate folder so git can get the relevant info.
    """
    origin = Path().absolute()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(origin)


git_info_cache = {}


# git_info is the most expensive part of the entire script - so we will have to parallelize
def git_info(file_paths, exclude=None, root=CWD):
    exclude_args = f"--invert-grep --grep={exclude}" if exclude else ''
    procs = []
    with chdir(root):
        for file_path in file_paths:
            if file_path in git_info_cache:
                procs.append((file_path, git_info_cache[file_path]))
            else:
                args = (
                    f"echo $(git log -n 2 --pretty='format:{BLUE} %as/%an/%ae' {exclude_args} -- "
                    f"{file_path.replace(root, '')})")
                procs.append((file_path, subprocess.Popen(args, stdout=subprocess.PIPE, shell=True)))

        result = {}
        for file_path, proc in procs:
            if isinstance(proc, subprocess.Popen):
                lines = []
                for line in proc.stdout.readlines():
                    lines.append(line.decode("utf-8").strip())
                git_info_cache[file_path] = lines[0]
            result[file_path] = git_info_cache[file_path]
    return result


def walk_files(path):
    for root, _, files in os.walk(path):
        for file in files:
            yield os.path.join(root, file)


def build_index(conf_type, index_spec, root=CWD, teams=None):
    rel_path = os.path.join(root, "production", conf_type)
    teams = teams or {}
    index_table = {}
    for path in walk_files(rel_path):
        index_entry = build_entry(path, index_spec, conf_type, root=root, teams=teams)
        if index_entry is not None:
            index_table[index_entry["name"][0]] = index_entry
    return index_table


def find_string(text, word):
    start = text.find(word)
    while start > -1:
        yield start
        start = text.find(word, start + 1)


def highlight(text, word):
    result = ""
    prev_idx = 0
    for idx in find_string(text, word):
        result = result + text[prev_idx:idx] + HIGHLIGHT + word + NORMAL
        prev_idx = idx + len(word)
    result += text[prev_idx:len(text)]
    return result


def prettify_entry(entry, target, modification, show=10, root=CWD, trim_paths=False):
    lines = []
    if trim_paths:
        for field in filter(lambda x: x in entry, PATH_FIELDS):
            entry[field] = entry[field].replace(root, '')
    for column, values in entry.items():
        name = " "*(15 - len(column)) + column
        if column in FILTER_COLUMNS and len(values) > show:
            values = [value for value in set(values) if target in value]
            if (len(values) > show):
                truncated = ', '.join(values[:show])
                remaining = len(values) - show
                values = f"[{truncated} ... {GREY}{UNDERLINE}{remaining} more{NORMAL}]"
        if column == "file":
            values = f"{BOLD}{values} {modification}{NORMAL}"
        else:
            values = highlight(str(values), target)
        lines.append(f"{BOLD}{ORANGE}{name}{NORMAL} - {values}")
    content = "\n" + "\n".join(lines)
    return content


def find_in_index(index_table, target):
    def valid_entry(entry):
        return any([
            target in value
            for column, values in entry.items()
            if column in FILTER_COLUMNS
            for value in values
        ])
    return find_in_index_pred(index_table, valid_entry)


def find_in_index_pred(index_table, valid_entry):
    return [entry for _, entry in index_table.items() if valid_entry(entry)]


def display_entries(entries, target, root=CWD, trim_paths=False):
    git_infos = git_info([entry["file"] for entry in entries], root=root)
    display = []
    for entry in entries:
        info = git_infos[entry["file"]]
        pretty = prettify_entry(entry, target, info, root=root, trim_paths=trim_paths)
        display.append((info, pretty))

    for (_, pretty_entry) in sorted(display):
        print(pretty_entry)


def enrich_with_joins(gb_index, join_index, root=CWD, teams=None):
    # nested gb entries
    for _, join_entry in join_index.items():
        for gb in join_entry["_group_bys"]:
            entry = build_entry(gb, GB_INDEX_SPEC, "group_bys", root=root, teams=teams)
            gb_index[entry["name"][0]] = entry
    # lineage -> reverse index from gb -> join
    for _, group_by in gb_index.items():
        group_by["joins"] = []
        group_by["join_event_driver"] = []
    for _, join in join_index.items():
        for gb_name in join["group_bys"]:
            if gb_name in gb_index:
                gb_index[gb_name]["joins"].append(join["name"][0])
                if len(join["_events_driver"]) > 0:
                    gb_index[gb_name]["join_event_driver"].append(join["_events_driver"][0])


# reuse `git log` command result
file_to_author = {}
# extract information based on GROUPBY_INDEX_SPEC into this
gb_index = []
# extract information based on JOIN_INDEX_SPEC into this
join_index = []


def author_name_email(file, exclude=None):
    if not os.path.exists(file):
        return ("", "")
    if file not in file_to_author:
        for file, auth_str in git_info([file], exclude).items():
            file_to_author[file] = auth_str.split("/")[-2:]
    return file_to_author[file]


def conf_file(conf_type, conf_name):
    path_parts = ["production", conf_type]
    path_parts.extend(conf_name.split(".", 1))
    return os.path.join(*path_parts)


# args[0] is output tsv file
# args[1] is commit messages to exclude when extracting author and email information
def events_without_topics(output_file=None, exclude_commit_message=None):
    result = []
    emails = set()

    def is_events_without_topics(entry):
        found = len(entry["_event_topics"]) == 0 and len(entry["_event_tables"]) > 0
        is_online = len(entry["online"]) > 0
        joins = ", ".join(entry["joins"]) if len(entry["joins"]) > 0 else "STANDALONE"
        if found:
            file = entry["json_file"] if os.path.exists(entry["json_file"]) else entry["file"]
            producer_name, producer_email = author_name_email(file, exclude_commit_message)
            emails.add(producer_email)
            consumers = set()
            for join in entry["joins"]:
                conf_file_path = conf_file("joins", join)
                consumer_name, consumer_email = author_name_email(conf_file_path, exclude_commit_message)
                consumers.add(consumer_name)
                emails.add(consumer_email)
            row = [
                entry["name"][0],
                producer_name,
                is_online,
                entry["_event_tables"][0],
                joins,
                ", ".join(consumers)
            ]
            result.append(row)
        return found

    find_in_index_pred(gb_index, is_events_without_topics)
    if output_file:
        with open(os.path.expanduser(output_file), 'w') as tsv_file:
            for row in result:
                tsv_file.write('\t'.join(map(str, row))+'\n')
        print("wrote information about cases where events us used " +
              f"without topics set into file {os.path.expanduser(output_file)}")
    else:
        for row in result:
            print('\t'.join(map(str, row))+'\n')
    print(",".join(list(emails)))


def load_team_data(path):
    with open(path, 'r') as infile:
        teams = json.load(infile)
    base_defaults = teams.get('default', {})
    full_info = teams.copy()
    for team, values in teams.items():
        full_info[team] = dict(base_defaults, **values)
    return full_info


# register all handlers here
handlers = {
    "_events_without_topics": events_without_topics
}

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Explore tool for chronon")
    parser.add_argument("keyword", help="Keyword to look up keys")
    parser.add_argument("--conf-root", help="Conf root for the configs", default=CWD)
    parser.add_argument(
        "--handler-args", nargs="*", help="Special arguments for handler keywords of the form param=value")
    args = parser.parse_args()
    root = args.conf_root
    if not (root.endswith("chronon") or root.endswith("zipline")):
        print("This script needs to be run from chronon conf root - with folder named 'chronon' or 'zipline', found: "
              + root)
    teams = load_team_data(os.path.join(root, 'teams.json'))
    gb_index = build_index("group_bys", GB_INDEX_SPEC, root=root, teams=teams)
    join_index = build_index("joins", JOIN_INDEX_SPEC, root=root, teams=teams)
    enrich_with_joins(gb_index, join_index, root=root, teams=teams)

    candidate = args.keyword
    if candidate in handlers:
        print(f"{candidate} is a registered handler")
        handler = handlers[candidate]
        handler_args = {}
        for arg in args.handler_args:
            splits = arg.split("=", 1)
            assert len(splits) == 2, f"need args to handler for the form, param=value. Found and invalid arg:{arg}"
            key, value = splits
            handler_args[key] = value
        handler(**handler_args)
    else:
        group_bys = find_in_index(gb_index, args.keyword)
        display_entries(group_bys, args.keyword, root=root, trim_paths=True)




#     Copyright (C) 2023 The Chronon Authors.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import os
import re
from setuptools import find_packages, setup

current_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(current_dir, "README.md"), "r") as fh:
    long_description = fh.read()


with open(os.path.join(current_dir, "requirements/base.in"), "r") as infile:
    basic_requirements = [line for line in infile]


__version__ = "local"
__branch__ = "master"
def get_version():
    version_str = os.environ.get("CHRONON_VERSION_STR", __version__)
    branch_str = os.environ.get("CHRONON_BRANCH_STR", __branch__)
    # Replace "-SNAPSHOT" with ".dev"
    version_str = version_str.replace("-SNAPSHOT", ".dev")
    # If the prefix is the branch name, then convert it as suffix after '+' to make it Python PEP440 complaint
    if version_str.startswith(branch_str + "-"):
        version_str = "{}+{}".format(
            version_str.replace(branch_str + "-", ""),
            branch_str
        )

    # Replace multiple continuous '-' or '_' with a single period '.'.
    # In python version string, the label identifier that comes after '+', is all separated by periods '.'
    version_str = re.sub(r'[-_]+', '.', version_str)

    return version_str

setup(
    classifiers=[
        "Programming Language :: Python :: 3.7"
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    scripts=['ai/chronon/repo/explore.py', 'ai/chronon/repo/compile.py', 'ai/chronon/repo/run.py'],
    description="Chronon python API library",
    include_package_data=True,
    install_requires=basic_requirements,
    name="chronon-ai",
    packages=find_packages(),
    extras_require={
        # Extra requirement to have access to cli commands in python2 environments.
        "pip2compat": ["click<8"]
    },
    python_requires=">=3.7",
    url=None,
    version=get_version(),
    zip_safe=False
)




#     Copyright (C) 2023 The Chronon Authors.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from ai.chronon.repo.run import retry_decorator

import pytest


def generator():
    """ Simple generator to have a changing variable. """
    for i in range(10):
        yield i


@pytest.fixture
def generated():
    """ Build fixture for generator so it can increment on subsequent calls """
    return generator()


def fail_x_times(generated, x):
    """ Function that will fail x times before succeeding given the same generator. """
    y = next(generated)
    if y < x:
        raise ValueError(f"Value was {y}")
    return y


@retry_decorator(1, 1)
def fail_x_times_and_succeed(generated, x):
    """ Decorated function that will succeed after failing once. """
    fail_x_times(generated, x)


def test_retry(generated):
    with pytest.raises(Exception):
        fail_x_times(generated, 3)
    fail_x_times_and_succeed(generated, 3)

